from . basic import check, addflag, delflag, BIT


class BitFlag(object):
    def __init__(self, flags, value=0):
        if not isinstance(flags, list) and not isinstance(flags, tuple):
            raise TypeError('flags must be a list/tuple')
        if len(flags) == 0:
            raise ValueError('flags must have at least one string')
        self._val = {}
        if int(value) < 0:
            raise ValueError('value must be > 0')
        self._int = int(value)
        tmp = dir(self)
        if len(flags) > 64:
            raise OverflowError('max 64 flags, got {}'.format(len(flags)))
        for i in range(len(flags)):
            if flags[i] in tmp:
                raise ValueError('Flag \'{}\' has the same name as BitFlag method'.format(flags[i]))
            if '_' in (flags[i][0], flags[i][-1]):
                raise ValueError('Flag \'{}\' cannot start/end with underscore'.format(flags[i]))
        self._flags = flags

    def __getattribute__(self, attr):
        if attr[0] == '_' or attr not in self._flags:
            return super().__getattribute__(attr)
        return check(self._int, self._get_bit(attr))

    def __setattr__(self, attr, value):
        if attr[0] == '_':
            return super().__setattr__(attr, value)
        if value not in (True, False):
            raise ValueError('Flag \'{}\' can only be True or False, got \'{}\''.format(attr, value))
        if value == False:
            self._int = delflag(self._int, self._get_bit(attr))
        else:
            self._int = addflag(self._int, self._get_bit(attr))

    def __setitem__(self, key, value):
        if value not in (True, False):
            raise ValueError('Flag \'{}\' can only be True or False, got \'{}\''.format(key, value))
        if value == False:
            self._int = delflag(self._int, self._get_bit(key))
        else:
            self._int = addflag(self._int, self._get_bit(key))

    def __getitem__(self, key):
        return check(self._int, self._get_bit(key))

    def _get_bit(self, key):
        if key not in self._flags:
            raise KeyError('`{}` not in flag list'.format(key))
        return BIT[self._flags.index(key)]

    def as_int(self):
        return self._int

    def as_dict(self):
        return {k:check(self._int, self._get_bit(k)) for k in self._flags}

    def as_list(self):
        return list(self.as_dict().items())

    def compare(self, value):
        value = int(value)
        if int(value) < 0:
            raise ValueError('value must be > 0')
        return {k:check(value, self._get_bit(k)) for k in self._flags
                if check(self._int, self._get_bit(k)) != check(value, self._get_bit(k))}

    def update(self, value):
        r = self.compare(value)
        self._int = value
        return r


from functools import reduce


BIT = [1 << x for x in range(64)]
MAX = [reduce(lambda x,y: x|y, BIT[:i+1]) for i in range(len(BIT))]


def check(haystack, needle):
    if isinstance(needle, list) or isinstance(needle, tuple):
        return check_multi(haystack, needle)
    return haystack & needle == needle

def check_multi(haystack, needles):
    i = newflag(needles)
    return haystack & i == i

def newflag(flags):
    return reduce(lambda x,y: x|y, flags)

def addflag(ori, flag):
    if isinstance(flag, list) or isinstance(flag, tuple):
        return addflag_multi(ori, flag)
    return ori | flag

def addflag_multi(ori, flags):
    i = newflag(flags)
    return ori | i

def delflag(ori, flag):
    if isinstance(flag, list) or isinstance(flag, tuple):
        return delflag_multi(ori, flag)
    return ori ^ flag if addflag(ori, flag) == ori else ori

def delflag_multi(ori, flags):
    tmp = ori
    for flag in flags:
        tmp = delflag(tmp, flag)
    return tmp

def bitcount(num):
    bit = 0
    while num > 0:
        bit += 1
        num >>= 1
    return bit

def prepend_num(initial, length, value):
    return (initial << length) | value

def explode_num(num, lengths):
    result = []
    for length in lengths:
        result.append(num & MAX[length-1])
        num >>= length
    return result

def implode_num(data):
    result = 0
    for length, value in data[::-1]:
        result = prepend_num(result, length, value)
    return result


from . basic import bitcount, explode_num, implode_num


class BitValue(object):
    def __init__(self, vals, num=0):
        if not isinstance(vals, dict):
            raise TypeError('`vals` must be a dict')
        if len(vals) == 0:
            raise ValueError('vals must have at lease one key:value')
        self._val = {k:dict(length=bitcount(v), max=v, value=0) for k,v in vals.items()}
        if sum([x['length'] for x in self._val.values()]) > 64:
            raise OverflowError('max length is 64 bits, got {}'.format(sum([x['length'] for x in self._val.values()])))
        if num > 0:
            t = explode_num(num, [x['length'] for x in self._val.values()])
            for k,v in zip(self._val.keys(), t):
                self._val[k]['value'] = v

    def __setitem__(self, key, value):
        if value > self._val[key]['max']:
            raise ValueError('Max value for {} is {} got {}'.format(key,
                             self._val[key]['max'], value))
        elif value < 0:
            raise ValueError('Min value is 0')
        self._val[key]['value'] = value

    def __getitem__(self, key):
        return self._val[key]['value']

    def __setattr__(self, key, value):
        if key[0] == '_':
            return super().__setattr__(key, value)
        if value > self._val[key]['max']:
            raise ValueError('max value for {} is {}, got {}'.format(key,
                             self._val[key]['max'], value))
        self._val[key]['value'] = value

    def __getattribute__(self, key):
        if key[0] == '_' or key not in self._val:
            return super().__getattribute__(key)
        return self._val[key]['value']

    def as_int(self):
        return implode_num([(v['length'],v['value']) for v in self._val.values()])

    def as_dict(self):
        return {k:v['value'] for k,v in self._val.items()}

    def as_list(self):
        return [(k,v['value']) for k,v in self._val.items()]

    def compare(self, value):
        value = int(value)
        if value < 0:
            raise ValueError('value must be > 0')
        t = explode_num(value, [x['length'] for x in self._val.values()])
        return {k:v for k,v in zip(self._val.keys(), t) if v != self._val[k]['value']}

    def update(self, value):
        value = int(value)
        r = self.compare(value)
        for k,v in r.items():
            self._val[k]['value'] = v
        return r